"""Module containing methods to build a shapegrid."""
import math

import numpy as np
from osgeo import ogr, osr

# Calculate this once and store as a constant instead of for every cell
SQRT_3 = math.sqrt(3)


# .............................................................................
def make_polygon_wkt_from_points(points):
    """Create a polygon WKT string from a list of points.

    Args:
        points (list of tuple): A list of (x,y) points for each vertex of the polygon.

    Note:
        Points should be ordered following right hand rule.

    Returns:
        str: A polygon well-known text string.
    """
    points.append(points[0])
    point_strings = ['{} {}'.format(x, y) for x, y in points]
    return 'POLYGON(({}))'.format(','.join(point_strings))


# .............................................................................
def hexagon_wkt_generator(min_x, min_y, max_x, max_y, x_res, y_res):
    """Generator producing hexagonal WKT for cells of the shapegrid.

    Args:
        min_x (numeric): The minimum x value.
        min_y (numeric): The minimum y value.
        max_x (numeric): The maximum x value.
        max_y (numeric): The maximum y value.
        x_res (numeric): The x size of the cell.
        y_res (numeric): The y size of the cell.

    Yields:
        str: Well-known text polygons for the cells.
    """
    y_coord = max_y
    y_row = True
    y_step = -1 * y_res * SQRT_3 / 4
    x_step = x_res * 1.5

    while y_coord > min_y:
        # Every other row needs to be shifted slightly
        x_coord = min_x if y_row else min_x + (x_res * .75)

        while x_coord < max_x:
            yield make_polygon_wkt_from_points([
                (x_coord - (x_res * .25), y_coord + (y_res * .25) * SQRT_3),
                (x_coord + (x_res * .25), y_coord + (y_res * .25) * SQRT_3),
                (x_coord + (x_res * .25), y_coord),
                (x_coord + (x_res * .25), y_coord - (y_res * .25) * SQRT_3),
                (x_coord - (x_res * .25), y_coord - (y_res * .25) * SQRT_3),
                (x_coord - (x_res * .5), y_coord)])
            x_coord += x_step
        y_row = not y_row
        y_coord += y_step


# .............................................................................
def square_wkt_generator(min_x, min_y, max_x, max_y, x_res, y_res):
    """Generator producing square WKT for cells of the shapegrid.

    Args:
        min_x (numeric): The minimum x value.
        min_y (numeric): The minimum y value.
        max_x (numeric): The maximum x value.
        max_y (numeric): The maximum y value.
        x_res (numeric): The x size of the cell.
        y_res (numeric): The y size of the cell.

    Yields:
        str: Well-known text polygons for the cells.
    """
    for y_coord in np.arange(max_y, min_y, -y_res):
        for x_coord in np.arange(min_x, max_x, x_res):
            yield make_polygon_wkt_from_points([
                (x_coord, y_coord),
                (x_coord + x_res, y_coord),
                (x_coord + x_res, y_coord - y_res),
                (x_coord, y_coord - y_res)])


# .............................................................................
def build_shapegrid(
    shapegrid_file_name,
    min_x,
    min_y,
    max_x,
    max_y,
    cell_size,
    epsg_code,
    cell_sides,
    site_id='siteid',
    site_x='siteX',
    site_y='siteY',
    cutout_wkt=None
):
    """Build a shapegrid with an optional cutout.

    Args:
        shapegrid_file_name (str): The location to store the resulting
            shapegrid.
        min_x (numeric): The minimum value for X of the shapegrid.
        min_y (numeric): The minimum value for Y of the shapegrid.
        max_x (numeric): The maximum value for X of the shapegrid.
        max_y (numeric): The maximum value for Y of the shapegrid.
        cell_size (numeric): The size of each cell (in units indicated by EPSG).
        epsg_code (int): The EPSG code for the new shapegrid.
        cell_sides (int): The number of sides for each cell of the shapegrid.
            4 - square cells, 6 - hexagon cells
        site_id (str): The name of the site id field for the shapefile.
        site_x (str): The name of the X field for the shapefile.
        site_y (str): The name of the Y field for the shapefile.
        cutout_wkt (None or str): WKT for an area of the shapegrid to be cut
            out.

    Returns:
        int: The number of cells in the new shapegrid.

    Raises:
        ValueError: Raised if an invalid number of cell sides is specified.
    """
    # We'll always check for intersection to reduce amount of work
    if cutout_wkt is None:
        cutout_wkt = make_polygon_wkt_from_points([
            (min_x, max_y), (max_x, max_y), (max_x, min_y), (min_x, min_y)])
    selected_poly = ogr.CreateGeometryFromWkt(cutout_wkt)

    # Just in case we decide that these can vary at some point
    x_res = y_res = cell_size

    # Initialize shapefile
    target_srs = osr.SpatialReference()
    target_srs.ImportFromEPSG(epsg_code)

    drv = ogr.GetDriverByName('ESRI Shapefile')
    data_set = drv.CreateDataSource(shapegrid_file_name)

    layer = data_set.CreateLayer(
        data_set.GetName(), geom_type=ogr.wkbPolygon, srs=target_srs)
    layer.CreateField(ogr.FieldDefn(site_id, ogr.OFTInteger))
    layer.CreateField(ogr.FieldDefn(site_x, ogr.OFTReal))
    layer.CreateField(ogr.FieldDefn(site_y, ogr.OFTReal))

    # Set up generator
    if cell_sides == 4:
        wkt_generator = square_wkt_generator(
            min_x, min_y, max_x, max_y, x_res, y_res)
    elif cell_sides == 6:
        wkt_generator = hexagon_wkt_generator(
            min_x, min_y, max_x, max_y, x_res, y_res)
    else:
        raise ValueError(
            'Cannot generate shapegrid cells with {} sides'.format(cell_sides)
        )

    shape_id = 0
    for cell_wkt in wkt_generator:
        geom = ogr.CreateGeometryFromWkt(cell_wkt)
        # Check for intersection
        if geom.Intersection(selected_poly):
            feat = ogr.Feature(feature_def=layer.GetLayerDefn())
            feat.SetGeometry(geom)
            centroid = geom.Centroid()
            feat.SetField(site_x, centroid.GetX())
            feat.SetField(site_y, centroid.GetY())
            feat.SetField(site_id, shape_id)
            layer.CreateFeature(feat)
            shape_id += 1
        feat.Destroy()
    data_set.Destroy()
    return shape_id


# .............................................................................
__all__ = [
    'build_shapegrid',
    'hexagon_wkt_generator',
    'make_polygon_wkt_from_points',
    'square_wkt_generator'
]



"""Module containing Point class.

Note: A namedtuple could replace this class for Python 3.7+
"""
import copy
import csv
import io
import json
import zipfile

import defusedxml.ElementTree as ET


DEFAULT_META_FILENAME = 'meta.xml'
# Metadata about occurrence records in DWCA with default values
# From https://dwc.tdwg.org/text/tdwg_dwc_text.xsd
DWCA_OCCURRENCE_PARAMS = {
    # Key: Default
    'linesTerminatedBy': '\n',
    'fieldsTerminatedBy': ',',
    'fieldsEnclosedBy': '"',
    'ignoreHeaderLines': 0,
    'rowType': None,  # Required
    'encoding': 'UTF-8',
    'dateFormat': 'YYY-MM-DD',
}

DEFAULT_SPECIES_TERM = 'scientificName'
DEFAULT_X_TERM = 'decimalLongitude'
DEFAULT_Y_TERM = 'decimalLatitude'

# DWCA Tag Constants
CORE_TAG = '{http://rs.tdwg.org/dwc/text/}core'
FIELD_TAG = '{http://rs.tdwg.org/dwc/text/}field'
FILES_TAG = '{http://rs.tdwg.org/dwc/text/}files'
ID_TAG = '{http://rs.tdwg.org/dwc/text/}id'
LOCATION_TAG = '{http://rs.tdwg.org/dwc/text/}location'
# EXTENSION_TAG = '{http://rs.tdwg.org/dwc/text/}extension'

# ROW_TYPE_ATT = 'rowType'
OCCURRENCE_ROW_TYPE = 'http://rs.tdwg.org/dwc/terms/Occurrence'


# .....................................................................................
class Point:
    """Class representing an occurrence data point."""
    # .......................
    def __init__(self, species_name, x, y, attributes=None):
        """Constructor.

        Args:
            species_name (:obj:`str`): The species name for this point.
            x (:obj:`float`): The value of the x coordinate for this occurrence point.
            y (:obj:`float`): The value of the y coordinate for this occurrence point.
            attributes (:obj:`dict`): A dictionary of attributes associated with this
                point.

        Raises:
            ValueError: Raised if the species name is omitted.
        """
        if species_name is None or len(species_name) < 1:
            raise ValueError('Species name must be provided')
        self.species_name = species_name.capitalize()
        self.x = float(x)
        self.y = float(y)
        if attributes is None:
            self.attributes = {}
        else:
            self.attributes = attributes
        # Set attributes for species name, x, and y
        self.attributes['species_name'] = self.species_name
        self.attributes['x'] = self.x
        self.attributes['y'] = self.y

    # .......................
    def __eq__(self, other):
        """Test if this point equals the other.

        Args:
            other (:obj:`Point`): A different Point object to compare with.

        Returns:
            bool: An indication if the two points are equal for the primary attributes.
        """
        return self.species_name == other.species_name and \
            self.x == other.x and self.y == other.y

    # .......................
    def __lt__(self, other):
        """Test if this point is less than the other.

        Args:
            other (:obj:`Point`): A different Point object to compare with.

        Returns:
            bool: An indication if this point is less than the other for the primary
                attributes.
        """
        if self.species_name < other.species_name:
            return True
        if self.species_name == other.species_name:
            if self.x < other.x:
                return True
            if self.x == other.x:
                return self.y < other.y
        return False

    # .......................
    def __repr__(self):
        """Get a string representation of this Point object.

        Returns:
            str: A string representation of this Point.
        """
        return 'Point(species="{}", x={}, y={})'.format(
            self.species_name, self.x, self.y)

    # .......................
    def get_attribute(self, attribute_name):
        """Get an attribute for the point.

        Args:
            attribute_name (:obj:`str`): The attribute to attempt to retrieve.

        Returns:
            object: The value of the attribute if it exists.
            None: Returned if the attribute does not exist for the Point.
        """
        if attribute_name in self.attributes.keys():
            return self.attributes[attribute_name]
        return None

    # .......................
    def set_attribute(self, attribute_name, value):
        """Set an attribute for the point.

        Args:
            attribute_name (:obj:`str`): The name of the attribute to set.
            value (:obj:`object`): The value to set the attribute to.
        """
        self.attributes[attribute_name] = value


# .....................................................................................
class PointCsvReader:
    """Class for reading Points from a CSV file."""
    # .......................
    def __init__(
        self,
        filename,
        species_field,
        x_field,
        y_field,
        geopoint=None,
        group_field='species_name'
    ):
        """Constructor for a Point CSV retriever.

        Args:
            filename (:obj:`str`): A file path containing CSV occurrence data.
            species_field (:obj:`str`): The field name of the column containing species
                data.
            x_field (:obj:`str`): The field name of the column containing x
                coordinates.
            y_field (:obj:`str`): The field name of the column containing y
                coordinates.
            geopoint (:obj:`str`): The field name of the column containing geopoint
                data.
            group_field (:obj:`str`): The name of the field to use for grouping points.
        """
        self.filename = filename
        self.file = None
        self.reader = None
        self.species_field = species_field
        self.x_field = x_field
        self.y_field = y_field
        self.geopoint = geopoint
        self.group_field = group_field
        self._next_points = []
        self._curr_val = None

    # .......................
    def __enter__(self):
        """Context manager magic method.

        Returns:
            PointCsvReader: This instance.
        """
        self.open()
        return self

    # .......................
    def __exit__(self, *args):
        """Context manager magic method on exit.

        Args:
            *args: Positional arguments passed to the exit function.
        """
        self.close()

    # .......................
    def __iter__(self):
        """Iterator magic method.

        Returns:
            PointCsvReader: This instance.
        """
        return self

    # .......................
    def __next__(self):
        """Get lists of consecutive points with the same attribute value.

        Returns:
            list: A list of point objects.

        Raises:
            KeyError: Raised if an attribute is missing.
            StopIteration: Raised when there are no additional objects.
        """
        for point_dict in self.reader:
            try:
                if self.geopoint is not None:
                    x_val = json.loads(point_dict[self.geopoint])[self.x_field]
                    y_val = json.loads(point_dict[self.geopoint])[self.y_field]
                else:
                    x_val = point_dict[self.x_field]
                    y_val = point_dict[self.y_field]
                pt = Point(
                    point_dict[self.species_field], x_val, y_val,
                    attributes=point_dict)
                test_val = pt.get_attribute(self.group_field)
                if test_val != self._curr_val:
                    if self._curr_val is not None:
                        self._curr_val = test_val
                        tmp = self._next_points
                        self._next_points = [pt]
                        return tmp
                    self._curr_val = test_val
                self._next_points.append(pt)

            except ValueError:  # pragma: no cover
                pass
            except KeyError as ke:  # pragma: no cover
                raise ke
            except TypeError:  # pragma: no cover
                pass
        if self._next_points:
            tmp = self._next_points
            self._next_points = []
            return tmp
        raise StopIteration

    # .......................
    def open(self):
        """Open the file and initialize."""
        self.file = open(self.filename, 'r')
        temp_lines = [next(self.file), next(self.file), next(self.file)]
        dialect = csv.Sniffer().sniff('\n'.join(temp_lines), delimiters="\t,")
        self.file.seek(0)
        self.reader = csv.DictReader(self.file, dialect=dialect)

    # .......................
    def close(self):
        """Close the file."""
        self.file.close()


# .....................................................................................
class PointCsvWriter():
    """Class for writing Points to a CSV file."""
    # .......................
    def __init__(self, filename, fields, **kwargs):
        """Constructor for writing points to csv file.

        Args:
            filename (:obj:`str`): A file location to write points to.
            fields (:obj:`list`): A list of fields to include in the csv headers.
            **kwargs (:obj:`dict`): Keyword parameters that will be passed to the
                DictWriter instance from the csv module.
        """
        self.filename = filename
        self.file = None
        self.writer = None
        self.field_names = fields
        self.kwargs = kwargs

    # .......................
    def __enter__(self):
        """Context manager magic method.

        Returns:
            PointCsvWriter: This instance.
        """
        self.open()
        return self

    # .......................
    def __exit__(self, *args):
        """Context manager magic method on exit.

        Args:
            *args: Positional arguments passed to the exit function.
        """
        self.close()

    # .......................
    def close(self):
        """Close file."""
        self.file.close()

    # .......................
    def open(self):
        """Open file for writing."""
        self.file = open(self.filename, 'w')
        self.writer = csv.DictWriter(self.file, self.field_names, **self.kwargs)
        self.writer.writeheader()

    # .......................
    def write_points(self, points):
        """Write a Point object to the CSV file.

        Args:
            points (:obj:`list` of :obj:`Point`): A list of points to write.
        """
        if isinstance(points, Point):
            points = [points]

        for point in points:
            point_dict = {k: point.get_attribute(k) for k in self.field_names}
            self.writer.writerow(point_dict)


# .....................................................................................
class PointDwcaReader:
    """Class for reading Darwin Core Archives."""
    # .......................
    def __init__(
        self,
        dwca_filename,
        meta_filename=DEFAULT_META_FILENAME,
        species_term=DEFAULT_SPECIES_TERM,
        x_term=DEFAULT_X_TERM,
        y_term=DEFAULT_Y_TERM,
        geopoint_term=None
    ):
        """Constructor for reading Darwin Core Archives.

        Args:
            dwca_filename (:obj:`str`): File location of a DWCA zip file.
            meta_filename (:obj:`str`): File within the archive containing metadata.
                Defaults to DEFAULT_META_FILENAME.
            species_term (:obj:`str`): Species term in the DWCA file.  Defaults to
                DEFAULT_SPECIES_TERM.
            x_term (:obj:`str`): X term in the DWCA file.  Defaults to DEFAULT_X_TERM.
            y_term (:obj:`str`): Y term in the DWCA file.  Defaults to DEFAULT_Y_TERM.
            geopoint_term (:obj:`str`): Geopoint term in the DWCA file.
                Default is None.
        """
        self.meta_filename = meta_filename
        self.archive_filename = dwca_filename
        self.occurrence_filename = None
        self.fields = {}
        self.occurrence_params = copy.deepcopy(DWCA_OCCURRENCE_PARAMS)
        self._curr_val = None
        self._next_points = []
        self.species_term = species_term
        self.x_term = x_term
        self.y_term = y_term
        self.geopoint_term = geopoint_term
        self.group_field = self.species_term

    # .......................
    def _get_species_name(self, point_dict):
        """Get the species name from the attribute dictionary.

        Args:
            point_dict (:obj:`dict`): A dictionary of point attributes.

        Returns:
            str: A species name
        """
        return point_dict[self.species_term]

    # .......................
    def _get_x_value(self, point_dict):
        """Get the x coordinate value from the attribute dictionary.

        Args:
            point_dict (:obj:`dict`): A dictionary of point attributes.

        Returns:
            numeric: The x coordinate retrieved.
            None: Returned if there is no x value.
        """
        if self.geopoint_term is not None:
            try:
                return json.loads(
                    point_dict[self.geopoint_term].replace("'", '"')
                )[self.x_term]
            except Exception:
                return None
        return point_dict[self.x_term]

    # .......................
    def _get_y_value(self, point_dict):
        """Get the y coordinate value from the attribute dictionary.

        Args:
            point_dict (:obj:`dict`): A dictionary of point attributes.

        Returns:
            numeric: The y coordinate retrieved.
            None: Returned if there is no y value.
        """
        if self.geopoint_term is not None:
            try:
                return json.loads(
                    point_dict[self.geopoint_term].replace("'", '"')
                )[self.y_term]
            except Exception:
                return None
        return point_dict[self.y_term]

    # .......................
    def __enter__(self):
        """Context manager magic method.

        Returns:
            PointDwcaReader: This instance.
        """
        self.open()
        return self

    # .......................
    def __exit__(self, *args):
        """Context manager magic method on exit.

        Args:
            *args: Positional arguments passed to the exit function.
        """
        self.close()

    # .......................
    def __iter__(self):
        """Iterator magic method.

        Returns:
            PointDwcaReader: This instance.
        """
        return self

    # .......................
    def __next__(self):
        """Get lists of consecutive points with the same attribute value.

        Returns:
            list: A list of point objects.

        Raises:
            StopIteration: Raised when there are no additional objects.
        """
        for point_row in self.reader:
            try:
                point_dict = {
                    term: self.fields[term](point_row) for term in self.fields.keys()
                }
                pt = Point(
                    self._get_species_name(point_dict),
                    self._get_x_value(point_dict),
                    self._get_y_value(point_dict),
                    attributes=point_dict
                )
                test_val = pt.get_attribute(self.group_field)
                if test_val != self._curr_val:
                    if self._curr_val is not None:
                        self._curr_val = test_val
                        tmp = self._next_points
                        self._next_points = [pt]
                        return tmp
                    self._curr_val = test_val
                self._next_points.append(pt)
            except KeyError:
                pass
            except TypeError:
                pass

        if self._next_points:
            tmp = self._next_points
            self._next_points = []
            return tmp
        raise StopIteration

    # .......................
    def _process_metadata(self, meta_contents):
        """Process the metadata file contained in the archive.

        Args:
            meta_contents (:obj:`str`): The string contents of the metadata file
                (meta.xml).
        """
        root_element = ET.fromstring(meta_contents)
        core_element = root_element.find(CORE_TAG)

        # If core element is missing (iDigBio) look in extensions
        # if core_element is None:
        #    for extension_el in root_element.findall(EXTENSION_TAG):
        #        if core_element is None and \
        #           extension_el.attrib[ROW_TYPE_ATT] == OCCURRENCE_ROW_TYPE:
        #            core_element = extension_el

        # Process core element
        # - Look for parameters we use for processing
        for core_att in self.occurrence_params.keys():
            if core_att in core_element.attrib.keys():
                self.occurrence_params[core_att] = core_element.attrib[core_att]

        # Get the occurrence data file name in the zip file
        self.occurrence_filename = core_element.find(FILES_TAG).findall(
            LOCATION_TAG
        )[0].text

        # Get the CSV fields from the metadata
        for field_element in core_element.findall(FIELD_TAG):
            # Get field processing function
            field_term = field_element.attrib['term']
            # Remove namespace
            if field_term.find('/') > 0:
                field_term = field_term.split('/')[-1]
            field_index = None
            field_default = None
            field_vocabulary = None
            field_delimiter = None

            if 'index' in field_element.attrib.keys():
                field_index = int(field_element.attrib['index'])
            if 'default' in field_element.attrib.keys():
                field_default = field_element.attrib['default']
            if 'vocabulary' in field_element.attrib.keys():
                field_vocabulary = field_element.attrib['vocabulary']
            if 'delimitedBy' in field_element.attrib.keys():
                field_delimiter = field_element.attrib['delimitedBy']

            self.fields[field_term] = get_field_process_func(
                index=field_index,
                default=field_default,
                vocabulary=field_vocabulary,
                delimiter=field_delimiter
            )
        # Check for id field
        for id_element in core_element.findall(ID_TAG):
            field_term = 'id'
            field_index = None
            field_default = None
            field_vocabulary = None
            field_delimiter = None

            if 'index' in id_element.attrib.keys():
                field_index = int(id_element.attrib['index'])
            if 'default' in id_element.attrib.keys():
                field_default = id_element.attrib['default']
            if 'vocabulary' in id_element.attrib.keys():
                field_vocabulary = id_element.attrib['vocabulary']
            if 'delimitedBy' in id_element.attrib.keys():
                field_delimiter = id_element.attrib['delimitedBy']

            self.fields[field_term] = get_field_process_func(
                index=field_index,
                default=field_default,
                vocabulary=field_vocabulary,
                delimiter=field_delimiter
            )

    # .......................
    def open(self):
        """Open the file and initialize."""
        # Open the zip file
        self._zip_archive = zipfile.ZipFile(self.archive_filename, mode='r')
        # self._zip_archive.open()
        meta_contents = io.TextIOWrapper(
            self._zip_archive.open(self.meta_filename)
        ).read()
        self._process_metadata(meta_contents)
        # Read metadata
        # Get occurrence file ready
        self.file = io.TextIOWrapper(
            self._zip_archive.open(self.occurrence_filename)
        )

        delimiter = self.occurrence_params['fieldsTerminatedBy']
        if delimiter.find('t') > 0:
            delimiter = '\t'

        reader_params = {
            'delimiter': delimiter,
        }
        if len(self.occurrence_params['linesTerminatedBy']) > 0:
            reader_params[
                'lineterminator'
            ] = self.occurrence_params['linesTerminatedBy']
        if len(self.occurrence_params['fieldsEnclosedBy']) > 0:
            reader_params['quotechar'] = self.occurrence_params['fieldsEnclosedBy']
        # reader_params['quotechar'] = None

        self.reader = csv.reader(self.file, **reader_params)
        for _ in range(int(self.occurrence_params['ignoreHeaderLines'])):
            next(self.reader)

    # .......................
    def close(self):
        """Close the file."""
        self.file.close()
        self._zip_archive.close()


# .....................................................................................
class PointJsonWriter():
    """Class for writing Points to JSON."""
    # .......................
    def __init__(self, filename):
        """Constructor for writing JSON points.

        Args:
            filename (:obj:`str`): A file location to write the points to.
        """
        self.filename = filename
        self.points = []

    # .......................
    def __enter__(self):
        """Context manager magic method.

        Returns:
            PointJsonWriter: This instance.
        """
        self.open()
        return self

    # .......................
    def __exit__(self, *args):
        """Exit and write JSON.

        Args:
            *args: Positional arguments sent to the exit function.
        """
        self.close()

    # .......................
    def close(self):
        """Close the writer."""
        with open(self.filename, 'w') as out_file:
            json.dump(self.points, out_file)

    # .......................
    def open(self):
        """Dummy method for consistency."""
        pass

    # .......................
    def write_points(self, points):
        """Add a point to the JSON output.

        Args:
            points (:obj:`list` of :obj:`Point`): A list of point objects to write out.
        """
        if isinstance(points, Point):
            points = [points]

        for point in points:
            self.points.append(point.attributes)


# .....................................................................................
def get_field_process_func(index=None, default=None, vocabulary=None, delimiter=None):
    """Get a function to process a field for a specimen record.

    Args:
        index (:obj:`int`, optional): The column index of the field value to process.
            If none, always return the default.
        default (number or string, optional): An optional default value (optional if
            index is not None) to return when the value of the field is empty.
        vocabulary (:obj:`str`, optional): A URI that identifies a vocabulary used for
            this field's possible values.
        delimiter (:obj:`str`, optional): An optional delimiter to split the field
            value with.

    Returns:
        Method: A method for getting the value of a field for a specimen row.
    """
    # .......................
    def default_getter(row):
        """Returns the default value, always.

        Args:
            row (:obj:`list`): A row of data for a specimen.

        Returns:
            object: Whatever the default value for the field is.
        """
        return default

    # .......................
    def list_getter(row):
        """Returns a list of value for the field.

        Args:
            row (:obj:`list`): A row of data for a specimen.

        Returns:
            list: A list of values generated by splitting the row index.
        """
        raw_val = ''
        if default is not None:
            raw_val = default
        if len(row[index]) > 0:
            raw_val = row[index]
        if len(raw_val) > 0:
            return raw_val.split(delimiter)
        return []

    # .......................
    def value_getter(row):
        """Returns a value for the field.

        Args:
            row (:obj:`list`): A row of data for a specimen.

        Returns:
            object: Whatever value is retrieved from the field.
        """
        if len(row[index]) > 0:
            if vocabulary is None or row[index] in vocabulary:
                return row[index]
        return default

    # Get the proper function
    if index is None:
        # Returned if there is no index to get data from, always return the default
        return default_getter
    if delimiter is not None:
        # If there is an index and a delimiter, return a list generating function
        return list_getter
    # If there is an index but no delimiter, return a field getter
    return value_getter


# .....................................................................................
def none_getter(obj):
    """Return None as a function.

    Args:
        obj (:obj:`object`): Any object.

    Returns:
        None: Always returns None.
    """
    return None



"""Data preparation module __init__."""
from . import build_grid
from . import layer_encoder
from . import occurrence_transformation

__all__ = []
#    'build_grid', 'layer_encoder', 'occurrence_filters',
#    'occurrence_transformation']
__all__.extend(build_grid.__all__)
__all__.extend(layer_encoder.__all__)
__all__.extend(occurrence_transformation.__all__)



"""Module containing base PAM statistic functionality."""
import numpy as np
from lmpy import Matrix


# .............................................................................
# Metric Decorators
# .............................................................................
class PamStatsMetric:
    """Base class for PAM statistics metrics.

    This wrapper is used to classify each statistic so that the arguments it
    needs can be inferred.
    """
    # .........................
    def __init__(self, func):
        """Constructor.

        Args:
            func: The function to call.
        """
        self.func = func
        self.__doc__ = self.func.__doc__

    # .........................
    def __call__(self, *args, **kwargs):
        """Call the wrapped function.

        Args:
            *args: Positional arguments passed to the function.
            **kwargs: Keyword arguments passed to the function.

        Returns:
            object: The output of the wrapped function.
        """
        # Do anything needed before calling the function
        ret = self.func(*args, **kwargs)
        # Do anything needed after the function
        return ret


# .............................................................................
class CovarianceMatrixMetric(PamStatsMetric):
    """A metric that produces a covariance matrix."""


# .............................................................................
class DiversityMetric(PamStatsMetric):
    """A single-value diversity metric."""


# .............................................................................
class _SiteStatMetric(PamStatsMetric):
    """A site-base statistic (do not use directly)."""


# .............................................................................
class _SpeciesStatMetric(PamStatsMetric):
    """A species-based statistic (do not use directly)."""


# .............................................................................
class SpeciesMatrixMetric(_SpeciesStatMetric):
    """A species-based statistic that is generated from a PAM."""


# .............................................................................
class SiteMatrixMetric(_SiteStatMetric):
    """A site-based metric that takes the PAM as input."""


# .............................................................................
class PamDistMatrixMetric(_SiteStatMetric):
    """A site-based metric computed from a PAM and Tree."""


# .............................................................................
class TreeMetric(_SiteStatMetric):
    """A site-based metric that takes a Dendropy tree as an argument."""


# .............................................................................
class TreeDistanceMatrixMetric(_SiteStatMetric):
    """A site-based metric that takes a distance matrix as an argument."""


# .............................................................................
# Site-based statistics
# .............................................................................
@SiteMatrixMetric
def alpha(pam):
    """Calculate alpha diversity, the number of species in each site.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        Matrix: A column of alpha diversity values for each site in the PAM.
    """
    return pam.sum(axis=1)


# .............................................................................
@SiteMatrixMetric
def alpha_proportional(pam):
    """Calculate proportional alpha diversity.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        Matrix: A column of proportional alpha diversity values for each site in the
            PAM.
    """
    return pam.sum(axis=1).astype(float) / num_species(pam)


# .............................................................................
@SiteMatrixMetric
def phi(pam):
    """Calculate phi, the range size per site.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        Matrix: A column of the sum of the range sizes for the species present at each
            site in the PAM.
    """
    return pam.dot(pam.sum(axis=0))


# .............................................................................
@SiteMatrixMetric
def phi_average_proportional(pam):
    """Calculate proportional range size per site.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        Matrix: A column of the proportional value of the sum of the range sizes for
            the species present at each site in the PAM.
    """
    return pam.dot(omega(pam)).astype(float) / (num_sites(pam) * alpha(pam))


# .............................................................................
# Species metrics
# .............................................................................
@SpeciesMatrixMetric
def omega(pam):
    """Calculate the range size per species.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        Matrix: A row of range sizes for each species in the PAM.
    """
    return pam.sum(axis=0)


# .............................................................................
@SpeciesMatrixMetric
def omega_proportional(pam):
    """Calculate the mean proportional range size of each species.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        Matrix: A row of the proportional range sizes for each species in the PAM.
    """
    return pam.sum(axis=0).astype(float) / num_sites(pam)


# .............................................................................
@SpeciesMatrixMetric
def psi(pam):
    """Calculate the range richness of each species.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        Matrix: A row of range richness for the sites that each species is present in.
    """
    return pam.sum(axis=1).dot(pam)


# .............................................................................
@SpeciesMatrixMetric
def psi_average_proportional(pam):
    """Calculate the mean proportional species diversity.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        Matrix: A row of proportional range richness for the sites that each species in
            the PAM is present.
    """
    return alpha(
        pam).dot(pam).astype(float) / (num_species(pam) * omega(pam))


# .............................................................................
# Diversity metrics
# .............................................................................
@DiversityMetric
def schluter_species_variance_ratio(pam):
    """Calculate Schluter's species variance ratio.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        float: The Schluter species variance ratio for the PAM.
    """
    sigma_species_ = sigma_species(pam)
    return float(sigma_species_.sum() / sigma_species_.trace())


# .............................................................................
@DiversityMetric
def schluter_site_variance_ratio(pam):
    """Calculate Schluter's site variance ratio.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        float: The Schluter site variance ratio for the PAM.
    """
    sigma_sites_ = sigma_sites(pam)
    return float(sigma_sites_.sum() / sigma_sites_.trace())


# .............................................................................
@DiversityMetric
def num_sites(pam):
    """Get the number of sites with presences.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        int: The number of sites that have present species.
    """
    return int(np.sum(np.any(pam, axis=1)))


# .............................................................................
@DiversityMetric
def num_species(pam):
    """Get the number of species with presences.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        int: The number of species that are present in at least one site.
    """
    return int(np.sum(np.any(pam, axis=0)))


# .............................................................................
@DiversityMetric
def whittaker(pam):
    """Calculate Whittaker's beta diversity metric for a PAM.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        float: Whittaker's beta diversity for the PAM.
    """
    return float(num_species(pam) / omega_proportional(pam).sum())


# .............................................................................
@DiversityMetric
def lande(pam):
    """Calculate Lande's beta diversity metric for a PAM.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        float: Lande's beta diversity for the PAM.
    """
    return float(num_species(pam) - (
        pam.sum(axis=0).astype(float) / num_sites(pam)).sum())


# .............................................................................
@DiversityMetric
def legendre(pam):
    """Calculate Legendre's beta diversity metric for a PAM.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        float: Legendre's beta diversity for the PAM.
    """
    return float(
        omega(pam).sum() - (float((omega(pam) ** 2).sum()) / num_sites(pam)))


# .............................................................................
@DiversityMetric
def c_score(pam):
    """Calculate the checker board score for the PAM.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        float: The checkerboard score for the PAM.
    """
    temp = 0.0
    # Cache these so we don't recompute
    omega_ = omega(pam)  # Cache so we don't waste computations
    num_species_ = num_species(pam)

    for i in range(num_species_):
        for j in range(i, num_species_):
            num_shared = len(np.where(np.sum(pam[:, [i, j]], axis=1) == 2)[0])
            p_1 = omega_[i] - num_shared
            p_2 = omega_[j] - num_shared
            temp += p_1 * p_2
    return 2 * temp / (num_species_ * (num_species_ - 1))


# .............................................................................
# Covariance metrics
# .............................................................................
@CovarianceMatrixMetric
def sigma_sites(pam):
    """Compute the site sigma metric for a PAM.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        Matrix: Matrix of covariance of composition of sites.
    """
    site_by_site = pam.dot(pam.T).astype(float)
    alpha_prop = alpha_proportional(pam)
    return (site_by_site / num_species(pam)) - np.outer(alpha_prop, alpha_prop)


# .............................................................................
@CovarianceMatrixMetric
def sigma_species(pam):
    """Compute the species sigma metric for a PAM.

    Args:
        pam (Matrix): The presence-absence matrix to use for the computation.

    Returns:
        Matrix: Matrix of covariance of composition of species.
    """
    species_by_site = pam.T.dot(pam).astype(float)
    omega_prop = omega_proportional(pam)
    return (species_by_site / num_sites(pam)
            ) - np.outer(omega_prop, omega_prop)


# .............................................................................
# Tree distance matrix metrics
# .............................................................................
@TreeDistanceMatrixMetric
def mean_nearest_taxon_distance(phylo_dist_mtx):
    """Calculates the nearest neighbor distance.

    Args:
        phylo_dist_mtx (Matrix): A matrix of distances between the species present at a
            site.

    Returns:
        float: The average distance from each taxa to taxa nearest to it.
    """
    try:
        nearest_total = np.sum(
            [np.min(row[row > 0.0]) for row in phylo_dist_mtx])
        return float(nearest_total / phylo_dist_mtx.shape[0])
    except Exception:  # pragma: no cover
        return 0.0


# .............................................................................
@TreeDistanceMatrixMetric
def mean_pairwise_distance(phylo_dist_mtx):
    """Calculates mean pairwise distance between species present at a site.

    Args:
        phylo_dist_mtx (Matrix): A matrix of distances between the species present at a
            site.

    Returns:
        float: The average distance from each taxa all of the other co-located taxa.
    """
    num_sp = phylo_dist_mtx.shape[0]
    return float((
        phylo_dist_mtx.sum() - phylo_dist_mtx.trace()
        ) / (num_sp * (num_sp - 1)))


# .............................................................................
@TreeDistanceMatrixMetric
def sum_pairwise_distance(phylo_dist_mtx):
    """Calculates the sum pairwise distance for all species present at a site.

    Args:
        phylo_dist_mtx (Matrix): A matrix of distances between the species present at a
            site.

    Returns:
        float: The total distance from each taxa all of the other co-located taxa.
    """
    return float((phylo_dist_mtx.sum() - phylo_dist_mtx.trace()) / 2.0)


# .............................................................................
@PamDistMatrixMetric
def pearson_correlation(pam, phylo_dist_mtx):
    """Calculates the Pearson correlation coefficient for each site.

    Args:
        pam (Matrix): A presence-absence matrix to use for the computation.
        phylo_dist_mtx (Matrix): A matrix of distance between species.

    Returns:
        Matrix: A column of Pearson correlation values for each site in a PAM.
    """
    num_sites_ = pam.shape[0]
    pearson = np.zeros((num_sites_, 1), dtype=float)

    for site_idx in range(num_sites_):
        sp_idxs = np.where(pam[site_idx] == 1)[0]
        num_sp = len(sp_idxs)
        num_pairs = num_sp * (num_sp - 1) / 2
        if num_pairs >= 2:
            # Need at least 2 pairs
            pair_dist = []
            pair_sites_shared = []
            for i in range(num_sp - 1):
                for j in range(i + 1, num_sp):
                    pair_dist.append(phylo_dist_mtx[i, j])
                    pair_sites_shared.append(
                        pam[:, i].dot(pam[:, j]))
            # X : Pair distance
            # Y : Pair sites shared
            x_val = np.array(pair_dist)
            y_val = np.array(pair_sites_shared)
            sum_xy = np.sum(x_val * y_val)
            sum_x = np.sum(x_val)
            sum_y = np.sum(y_val)
            sum_x_sq = np.sum(x_val ** 2)
            sum_y_sq = np.sum(y_val ** 2)

            # Pearson
            p_num = sum_xy - sum_x * sum_y / num_pairs
            p_denom = np.sqrt(
                (sum_x_sq - (sum_x ** 2 / num_pairs)
                 ) * (sum_y_sq - (sum_y ** 2 / num_pairs)))
            pearson[site_idx, 0] = p_num / p_denom
    return pearson


# .............................................................................
@TreeMetric
def phylogenetic_diversity(tree):
    """Calculate phylogenetic diversity of a tree.

    Args:
        tree (TreeWrapper): A phylogenetic tree to compute phylogenetic diversity for.

    Returns:
        float: The sum of the edge lengths of the nodes of the provided tree.
    """
    try:
        return np.sum([node.edge_length for node in tree.nodes()])
    except Exception:  # pragma: no cover
        return 0.0


# .............................................................................
class PamStats:
    """Class for managing metric computation for PAM statistics."""
    covariance_stats = [
        ('sigma sites', sigma_sites),
        ('sigma species', sigma_species)
        ]
    diversity_stats = [
        ('c-score', c_score),
        ('lande', lande),
        ('legendre', legendre),
        ('num sites', num_sites),
        ('num species', num_species),
        ('whittaker', whittaker)
        ]
    site_matrix_stats = [
        ('alpha', alpha),
        ('alpha proportional', alpha_proportional),
        ('phi', phi),
        ('phi average proportional', phi_average_proportional)
        ]
    site_tree_stats = [('Phylogenetic Diversity', phylogenetic_diversity)]
    site_tree_distance_matrix_stats = [
        ('MNTD', mean_nearest_taxon_distance),
        ('Mean Pairwise Distance', mean_pairwise_distance),
        ('Sum Pairwise Distance', sum_pairwise_distance)]
    site_pam_dist_mtx_stats = [('pearson_correlation', pearson_correlation)]
    species_matrix_stats = [
        ('omega', omega),
        ('omega_proportional', omega_proportional),
        ('psi', psi),
        ('psi_average_proportional', psi_average_proportional)
        ]

    # ...........................
    def __init__(self, pam, tree=None):
        """Constructor for PAM stats computations.

        Args:
            pam (Matrix): A presence-absence matrix to use for computations.
            tree (TreeWrapper): A tree to use for phylogenetic distance computations.
        """
        self.pam = pam
        self.tree = tree

    # ...........................
    def calculate_covariance_statistics(self):
        """Calculate covariance statistics matrices.

        Returns:
            list of tuple: A list of metric name, value tuples for covariance stats.
        """
        return [(name, func(self.pam)) for name, func in self.covariance_stats]

    # ...........................
    def calculate_diversity_statistics(self):
        """Calculate diversity statistics.

        Returns:
            list of tuple: A list of metric name, value tuples for diversity metrics.
        """
        print([func(self.pam) for _, func in self.diversity_stats])
        diversity_matrix = Matrix(
            np.array([func(self.pam) for _, func in self.diversity_stats]),
            headers={
                '0': ['value'],
                '1': [name for name, _ in self.diversity_stats]})
        return diversity_matrix

    # ...........................
    def calculate_site_statistics(self):
        """Calculate site-based statistics.

        Returns:
            Matrix: A matrix of site-based statistics for the selected metrics.
        """
        # Matrix based
        print('Start')
        site_stats_matrix = Matrix(
            np.zeros((self.pam.shape[0], len(self.site_matrix_stats))),
            headers={
                '0': self.pam.get_row_headers(),
                '1': [name for name, _ in self.site_matrix_stats]})
        print('Site matrix stats')
        for i in range(len(self.site_matrix_stats)):
            site_stats_matrix[:, i] = self.site_matrix_stats[i][1](self.pam)

        if self.tree is not None:
            print('Tree stuff')
            squid_annotations = self.tree.get_annotations('squid')
            squid_dict = {squid: label for label, squid in squid_annotations}
            ordered_labels = []
            for squid in self.pam.get_column_headers():
                if squid in squid_dict.keys():
                    v = squid_dict[squid]
                else:  # pragma: no cover
                    v = ''
                ordered_labels.append(v)

            site_tree_stats_matrix = Matrix(
                np.zeros((self.pam.shape[0], len(self.site_tree_stats))),
                headers={
                    '0': self.pam.get_row_headers(),
                    '1': [name for name, _ in self.site_tree_stats]})

            site_tree_dist_mtx_matrix = Matrix(
                np.zeros((self.pam.shape[0],
                          len(self.site_tree_distance_matrix_stats))),
                headers={
                    '0': self.pam.get_row_headers(),
                    '1': [name for name, _ in
                          self.site_tree_distance_matrix_stats]})

            # PAM / Tree stats
            print('Get distance matrix')
            phylo_dist_mtx = self.tree.get_distance_matrix()
            print('PAM dist mtx stats')
            site_pam_tree_matrix = None
            if self.site_pam_dist_mtx_stats:
                site_pam_tree_matrix = Matrix(
                    Matrix.concatenate(
                        [func(self.pam, phylo_dist_mtx
                              ) for _, func in self.site_pam_dist_mtx_stats]),
                    headers={
                        '0': self.pam.get_row_headers(),
                        '1': [name for name, _ in
                              self.site_pam_dist_mtx_stats]})

            print('Site by site')
            # Loop through PAM
            for site_idx, site_row in enumerate(self.pam):
                # Get present species
                present_species = np.where(site_row == 1)[0]

                # Get sub tree
                present_labels = list(
                    filter(lambda x: bool(x),
                           [ordered_labels[i] for i in present_species]))
                present_dist_mtx_idxs = []
                for idx, label in enumerate(
                        phylo_dist_mtx.get_column_headers()):
                    if label in present_labels:
                        present_dist_mtx_idxs.append(idx)
                try:
                    if present_labels:
                        site_tree = self.tree.extract_tree_with_taxa_labels(
                            present_labels)
                        # Get distance matrix
                        site_dist_mtx = phylo_dist_mtx.slice(
                            present_dist_mtx_idxs, present_dist_mtx_idxs)
                        # site_dist_mtx = site_tree.get_distance_matrix()
                        site_tree_dist_mtx_matrix[site_idx] = [
                            func(site_dist_mtx) for (
                                _, func
                                ) in self.site_tree_distance_matrix_stats]
                        site_tree_stats_matrix[site_idx] = [
                            func(site_tree
                                 ) for _, func in self.site_tree_stats]
                except Exception as err:  # pragma: no cover
                    print(err)
                    print(present_labels)
                    print('Site index: {}'.format(site_idx))

            all_stat_matrices = []
            if site_stats_matrix is not None:
                all_stat_matrices.append(site_stats_matrix)
            if site_tree_stats_matrix is not None:
                all_stat_matrices.append(site_tree_stats_matrix)
            if site_tree_dist_mtx_matrix is not None:
                all_stat_matrices.append(site_tree_dist_mtx_matrix)
            if site_pam_tree_matrix is not None:
                all_stat_matrices.append(site_pam_tree_matrix)

            site_stats_matrix = Matrix.concatenate(all_stat_matrices, axis=1)
        return site_stats_matrix

    # ...........................
    def calculate_species_statistics(self):
        """Calculate species-based statistics.

        Returns:
            Matrix: A matrix of species-based statistics for the selected metrics.
        """
        # Matrix based
        species_stats_matrix = Matrix(
            np.zeros((self.pam.shape[1], len(self.species_matrix_stats))),
            headers={
                '0': self.pam.get_column_headers(),
                '1': [name for name, _ in self.species_matrix_stats]})
        for i in range(len(self.species_matrix_stats)):
            species_stats_matrix[:, i] = \
                self.species_matrix_stats[i][1](self.pam)

        return species_stats_matrix

    # ...........................
    def register_metric(self, name, metric_function):
        """Register a new metric.

        Args:
            name (str): A name for this metric that will be used for a header.
            metric_function (function): A decorated metric generating function.

        Raises:
            TypeError: Raised if the metric function type cannot be processed.
        """
        if isinstance(metric_function, CovarianceMatrixMetric):
            self.covariance_stats.append((name, metric_function))
        elif isinstance(metric_function, DiversityMetric):
            self.diversity_stats.append((name, metric_function))
        elif isinstance(metric_function, SiteMatrixMetric):
            self.site_matrix_stats.append((name, metric_function))
        elif isinstance(metric_function, TreeMetric):
            self.site_tree_stats.append((name, metric_function))
        elif isinstance(metric_function, TreeDistanceMatrixMetric):
            self.site_tree_distance_matrix_stats.append(
                (name, metric_function))
        elif isinstance(metric_function, PamDistMatrixMetric):
            self.site_pam_dist_mtx_stats.append((name, metric_function))
        elif isinstance(metric_function, SpeciesMatrixMetric):
            self.species_matrix_stats.append((name, metric_function))
        else:
            raise TypeError(
                'Unknown metric type: {}, {}'.format(name, metric_function)
            )


# .............................................................................
__all__ = [
    'CovarianceMatrixMetric',
    'DiversityMetric',
    'PamDistMatrixMetric',
    'PamStats',
    'PamStatsMetric',
    'SiteMatrixMetric',
    'SpeciesMatrixMetric',
    'TreeDistanceMatrixMetric',
    'TreeMetric',
    'alpha',
    'alpha_proportional',
    'c_score',
    'lande',
    'legendre',
    'mean_nearest_taxon_distance',
    'mean_pairwise_distance',
    'num_sites',
    'num_species',
    'omega',
    'omega_proportional',
    'pearson_correlation',
    'phi',
    'phi_average_proportional',
    'phylogenetic_diversity',
    'psi',
    'psi_average_proportional',
    'schluter_site_variance_ratio',
    'schluter_species_variance_ratio',
    'sigma_sites',
    'sigma_species',
    'sum_pairwise_distance',
    'whittaker'
]



"""Module containing Gotelli swap method for randomizing a PAM."""
import numpy as np

from lmpy.matrix import Matrix

MAX_TRIES_WITHOUT_SWAP = 1000000


# .............................................................................
def swap_randomize(matrix, num_swaps, max_tries=MAX_TRIES_WITHOUT_SWAP):
    """Randomize a PAM matrix using the Gotelli swap method.

    Args:
        matrix (Matrix): A Lifemapper matrix object with binary elements
        num_swaps (int): The number of swaps to perform
        max_tries (int): The maximum number of attempts to swap before failing.

    Returns:
        Matrix: A randomized matrix.

    Raises:
        Exception: Raised when the maximum number of tries is reached without swapping.

    Todo:
        Allow num_swaps to be specified as a percentage of the matrix fill.
    """
    mtx_headers = matrix.get_headers()
    swapped_mtx = matrix.copy().astype(int)
    counter = 0
    num_tries = 0
    row_len, col_len = matrix.shape

    #  num_tries is a safety to kill the loop if nothing is ever found
    while counter < num_swaps and num_tries < max_tries:
        num_tries += 1
        column1 = np.random.randint(0, col_len)
        column2 = np.random.randint(0, col_len)
        row1 = np.random.randint(0, row_len)
        while column2 == column1:
            column2 = np.random.randint(0, col_len)
        first_corner = swapped_mtx[row1][column1]
        if first_corner ^ swapped_mtx[row1][column2]:
            row2 = np.random.randint(0, row_len)
            while row2 == row1:
                row2 = np.random.randint(0, row_len)
            if ((first_corner ^ swapped_mtx[row2][column1]) and
                    (not(first_corner) ^ swapped_mtx[row2][column2])):
                swapped_mtx[row1][column2] = first_corner
                swapped_mtx[row2][column1] = first_corner
                swapped_mtx[row2][column2] = not first_corner
                swapped_mtx[row1][column1] = not first_corner
                counter += 1
                num_tries = 0
    if num_tries >= max_tries:  # pragma: no cover
        raise Exception(
            'Reached maximum number of tries without finding suitable swap'
        )

    return Matrix(swapped_mtx, headers=mtx_headers)


# .............................................................................
def trial_swap(matrix, num_trials=None):
    """Randomize a PAM matrix using the trial swap method.

    Args:
        matrix (Matrix): A Lifemapper matrix object with binary elements.
        num_trials (int): The number of trials to perform.

    Returns:
        Matrix: A randomized matrix.

    Todo:
        Allow num_swaps to be specified as a percentage of the matrix fill
    """
    mtx_headers = matrix.get_headers()
    swapped_mtx = matrix.copy().astype(int)
    row_len, col_len = matrix.shape
    if num_trials is None:
        num_trials = matrix.size

    for _ in range(num_trials):
        column1 = np.random.randint(0, col_len)
        column2 = np.random.randint(0, col_len)
        row1 = np.random.randint(0, row_len)
        while column2 == column1:
            column2 = np.random.randint(0, col_len)
        first_corner = swapped_mtx[row1][column1]
        if first_corner ^ swapped_mtx[row1][column2]:
            row2 = np.random.randint(0, row_len)
            while row2 == row1:
                row2 = np.random.randint(0, row_len)
            if ((first_corner ^ swapped_mtx[row2][column1]) and
                    (not(first_corner) ^ swapped_mtx[row2][column2])):
                swapped_mtx[row1][column2] = first_corner
                swapped_mtx[row2][column1] = first_corner
                swapped_mtx[row2][column2] = not first_corner
                swapped_mtx[row1][column1] = not first_corner

    return Matrix(swapped_mtx, headers=mtx_headers)


# .............................................................................
__all__ = ['swap_randomize', 'trial_swap']



"""Module containing a class for encoding a Phylogenetic tree into a matrix.

See:
    Leibold, m.A., E.P. Economo and P.R. Peres-Neto. 2010. Metacommunity
        phylogenetics: separating the roles of environmental filters and
        historical biogeography. Ecology letters 13: 1290-1299.

Todo:
    Use method to get labels by matrix index when available from tree.
"""
from random import shuffle

import numpy as np
from lmpy import Matrix, PhyloTreeKeys, TreeWrapper


# .............................................................................
class EncodingException(Exception):
    """Exception indicating there was a problem with encoding."""


# .............................................................................
class TreeEncoder:
    """A class representing encoding of tree to match PAM."""

    # ..............................
    def __init__(self, tree, pam):
        """Base constructor for tree encoder.

        Args:
            tree (TreeWrapper): A tree object to encode.
            pam (Matrix): A PAM matrix object.
        """
        self.tree = tree

        if isinstance(pam, Matrix):
            self.pam = pam
        else:
            self.pam = Matrix(pam)

    # ..............................
    @classmethod
    def from_file(cls, tree_file_name, pam_file_name):
        """Creates an instance of the PhyloEncoding class from tree and pam.

        Args:
            tree_file_name (str): The location of the tree.
            pam_file_name (str): The location of the PAM.

        Returns:
            TreeEncoder: A new tree encoder instance.
        """
        tree = TreeWrapper.from_filename(tree_file_name)
        pam = Matrix.load(pam_file_name)
        return cls(tree, pam)

    # ..............................
    def encode_phylogeny(self):
        """Encode the phylogenetic tree into a matrix.

        Note:
            P in the literature, a tip (row) by internal node (column) matrix
                that needs to match the provided PAM.

        Todo:
            Consider providing options for correcting the tree / pam

        Raises:
            EncodingException: Raised if the PAM and tree do not match.

        Returns:
            Matrix: An encoding of the phylogenetic tree.
        """
        if self.validate():  # Make sure that the PAM and tree match
            if self.tree.has_branch_lengths():
                p_mtx = self._build_p_matrix_with_branch_lengths()
            else:
                p_mtx = self._build_p_matrix_no_branch_lengths()
        else:
            raise EncodingException(
                "PAM and Tree do not match, fix before encoding")

        return p_mtx

    # ..............................
    def validate(self):
        """Validates the tree / PAM combination.

        Returns:
            bool: Boolean indicating if the tree / pam is valid.
        """
        # check if tree is ultrametric
        if (not self.tree.has_branch_lengths() or self.tree.is_ultrametric()) \
                and self.tree.is_binary():
            # Check that matrix indices in tree match PAM
            # List of matrix indices (based on PAM column count)
            pam_mtx_indices = range(self.pam.shape[1])
            # All matrix indices in tree
            tree_mtx_indices = [
                int(mtx_id) for _, mtx_id in self.tree.get_annotations(
                    PhyloTreeKeys.MTX_IDX) if mtx_id is not None]

            # Find the intersection between the two lists
            intersection = set(pam_mtx_indices) & set(tree_mtx_indices)

            # This checks that there are no duplicates in either of the indices
            #     lists and that they overlap completely
            if len(intersection) == len(pam_mtx_indices) and \
                    len(pam_mtx_indices) == len(tree_mtx_indices):
                # If everything checks out to here, return true for valid
                return True
        # If anything does not validate, return false
        return False

    # ..............................
    def _build_p_branch_length_values(self, node):
        """Recurse through the tree to get P matrix values for node / tips.

        Args:
            node (dendropy.Node): The current clade.

        Returns:
            tuple: A tuple of branch length dictionary, sum of branch lengths, and
                p-values dictionary.
        """
        # Dictionary of branch length lists, bottom-up from tip to current node
        bl_dict = {}

        clade_bl = node.edge_length  # To avoid lookups

        # This is the sum of all branch lengths in the clade and will be used
        #    as the denominator for the p value for each tip to this node.  See
        #    the literature for more information.  We will initialize with the
        #    branch length of this clade because it will always be added
        #    whether we are at a tip or an internal node.
        bl_sum = clade_bl

        # Dictionary of dictionaries of p values, first key is node path id,
        #    sub-keys are tip matrix indices for that node
        p_vals_dict = {}

        if node.num_child_nodes() > 0:  # Assume this is two since binary
            clade_p_vals = {}
            multipliers = [-1.0, 1.0]  # One positive, one negative
            shuffle(multipliers)

            for child in node.child_nodes():
                (child_bl_dict, child_bl_sum, child_p_val_dict
                 ) = self._build_p_branch_length_values(child)
                multiplier = multipliers.pop(0)

                # Extend the p values dictionary
                p_vals_dict.update(child_p_val_dict)

                # Add this child's branch length sum to the clade's branch
                #    length sum
                bl_sum += child_bl_sum
                child_bl = child.edge_length

                # We will add this value to the branch length list for all of
                #    the tips in this clade.  It is the branch length of this
                #    clade divided by the number of tips in the clade.
                add_val = 1.0 * child_bl / len(child_bl_dict)

                # Process eac of the tips in the child_bl_dict
                for k, child_bl_k in child_bl_dict.items():
                    # The formula for calculating the p value is:
                    #    P(tip)(node) = (l1 + l2/2 + l3/3 + ... ln/n) / sum of
                    #        branch lengths to node
                    #    The value is arbitrarily set to be positive for one
                    #        child and negative for the other
                    #    The 'l' term is the length of a branch  and it is
                    #        divided by the number of tips that share that
                    #        branch.
                    tip_bls = child_bl_k + [add_val]

                    # Add the p value to p_vals_dict
                    clade_p_vals[k] = multiplier * sum(tip_bls) / child_bl_sum

                    # Add to bl_dict with this branch length
                    bl_dict[k] = tip_bls

            p_vals_dict[node.label] = clade_p_vals

        else:  # We are at a tip
            bl_dict = {
                node.taxon.annotations.get_value(PhyloTreeKeys.MTX_IDX): []
            }

        return bl_dict, bl_sum, p_vals_dict

    # ..............................
    def _build_p_matrix_no_branch_lengths(self):
        """Creates a P matrix when no branch lengths are present.

        Note:
            For this method, we assume that there is a total weight of -1 to
                the left and +1 to the right for each node.  As we go down
                (towards the tips) of the tree, we divide the proportion of
                each previously visited node by 2.  We then recurse with this
                new visited list down the tree.  Once we reach a tip, we can
                return that list of proportions because it will match for that
                tip for each of its ancestors.

        Example:
            3
            +-- 2
            |   +-- 1
            |   |   +-- 0
            |   |   |   +-- A
            |   |   |   +-- B
            |   |   |
            |   |   +-- C
            |   |
            |   +-- D
            |
            +--4
               +-- E
               +-- F

            Step 1: (Node 3) []
                - recurse left with [(3,-1)]
                - recurse right with [(3,1)]
            Step 2: (Node 2) [(3,-1)]
                - recurse left with [(3,-.5),(2,-1)]
                - recurse right with [(3,-.5),(2,1)]
            Step 3: (Node 1)[(3,-.5),(2,-1)]
                - recurse left with [(3,-.25),(2,-.5),(1,-1)]
                - recurse right with [3,-.25),(2,-.5),(1,1)]
            Step 4: (Node 0)[(3,-.25),(2,-.5),(1,-1)]
                - recurse left with [(3,-.125),(2,-.25),(1,-.5),(0,-1)]
                - recurse right with [(3,-.125),(2,-.25),(1,-.5),(0,1)]
            Step 5: (Tip A) - Return [(3,-.125),(2,-.25),(1,-.5),(0,-1)]
            Step 6: (Tip B) - Return [(3,-.125),(2,-.25),(1,-.5),(0,1)]
            Step 7: (Tip C) - Return [(3,-.25),(2,-.5),(1,1)]
            Step 8: (Tip D) - Return [(3,-.5),(2,1)]
            Step 9: (Node 4) [(3,1)]
                - recurse left with [(3,.5),(4,-1)]
                - recurse right with [(3,.5),(4,1)]
            Step 10: (Tip E) - Return [(3,.5),(4,-1)]
            Step 11: (Tip F) - Return [(3,.5),(4,1)]

            Creates matrix:
                     0     1     2      3       4
                A -1.0  -0.5  -0.25  -0.125   0.0
                B  1.0  -0.5  -0.25  -0.125   0.0
                C  0.0   1.0  -0.5   -0.25    0.0
                D  0.0   0.0   1.0   -0.5     0.0
                E  0.0   0.0   0.0    0.5    -1.0
                F  0.0   0.0   0.0    0.5     1.0

        See:
            Page 1293 of the literature.

        Returns:
            Matrix: An encoded phylogeny matrix.
        """
        # .......................
        # Initialize the matrix
        internal_node_labels = [
            n.label for n in self.tree.nodes() if not n.is_leaf()]
        labels = self.pam.get_column_headers()

        # We need a mapping of node path id to matrix column.  I don't think
        #     order matters
        node_col_idx = dict(
            zip(internal_node_labels, range(len(internal_node_labels))))

        mtx = Matrix(
            np.zeros((len(labels), len(internal_node_labels)), dtype=float),
            headers={'0': labels, '1': internal_node_labels})

        # Get the list of tip proportion lists
        # Note: Which tip each of the lists belongs to doesn't really matter
        #    but recursion will start at the top and go to the bottom of the
        #    tree tips.
        self.tree.seed_node._set_edge_length(0.0)
        tip_props = self._build_p_matrix_tip_proportion_list(
            self.tree.seed_node, visited=[])

        # The matrix index of the tip in the PAM maps to the row index of P
        for row_idx, tip_prop in tip_props:
            for node_clade_id, val in tip_prop:
                mtx[row_idx][node_col_idx[node_clade_id]] = val

        return mtx

    # ..............................
    def _build_p_matrix_tip_proportion_list(self, node, visited=None):
        """Builds a list of tip proportions for the p matrix w/o branch lengths.

        Args:
            node (dendropy.Node): The current clade.
            visited (:obj:`None` or :obj:`list` of :obj:`tuple`): A list of
                (node path id, proportion) tuples.

        Note:
            Proportion for each visited node is divided by two as we go
                towards the tips at each hop.

        Returns:
            list: A list of tip proportions.
        """
        if visited is None:
            visited = []  # pragma: no cover

        tip_props = []
        if node.num_child_nodes() > 0:  # Assume this is two since binary
            # First divide all existing visited by two
            new_visited = [(idx, val / 2) for idx, val in visited]
            # Recurse.  Left is negative, right is positive
            left_child, right_child = node.child_nodes()
            tip_props.extend(
                self._build_p_matrix_tip_proportion_list(
                    left_child, visited=new_visited + [(node.label, -1.0)]))
            tip_props.extend(
                self._build_p_matrix_tip_proportion_list(
                    right_child, visited=new_visited + [(node.label, 1.0)]))
        else:
            tip_props.append(
                (node.taxon.annotations.get_value(PhyloTreeKeys.MTX_IDX),
                 visited))

        return tip_props

    # ..............................
    def _build_p_matrix_with_branch_lengths(self):
        """Creates a P matrix when branch lengths are present.

        Note:
            For this method, we assume that there is a total weight of -1 to
                the left and +1 to the right for each node.  As we go down
                (towards the tips) of the tree, we divide the proportion of
                each previously visited node by 2.  We then recurse with this
                new visited list down the tree.  Once we reach a tip, we can
                return that list of proportions because it will match for that
                tip for each of its ancestors.

        Example:
            3
            +-- 2 (0.4)
            |    +-- 1 (0.15)
            |    |    +-- 0 (0.65)
            |    |    |    +-- A (0.2)
            |    |    |    +-- B (0.2)
            |    |    |
            |    |    +-- C (0.85)
            |    |
            |    +-- D (1.0)
            |
            +-- 4 (0.9)
                 +-- E (0.5)
                 +-- F (0.5)

            Value for any cell (tip)(node) = (l1 + l2/2 + l3/3 + ... + ln/n) /
                                   (Sum of branch lengths in daughter clade)
            ln / n -> The length of branch n divided by the number of tips that
                descend from that clade

            P(A)(2) = (0.2 + 0.65/2 + 0.15/3) /
                        (0.2 + 0.2 + 0.65 + 0.85 + 0.15)
                    = (0.2 + 0.325 + 0.05) / (2.05)
                    = 0.575 / 2.05
                    = 0.280

            P(D)(3) = (1.0 + 0.4/4) / (.2 + .2 + .65 + .85 + .15 + 1.0 + .4)
                    = 1.1 / 3.45
                    = 0.319

            Creates matrix:
                    0         1       2       3       4
                A  -1.000  -0.500  -0.280  -0.196   0.000
                B   1.000  -0.500  -0.280  -0.196   0.000
                C   0.000   1.000  -0.439  -0.290   0.000
                D   0.000   0.000   1.000  -0.319   0.000
                E   0.000   0.000   0.000   0.500  -1.000
                F   0.000   0.000   0.000   0.500   1.000

        See:
            Literature supplemental material.

        Returns:
            Matrix: An encoded phylogeny matrix.
        """
        # We only need the P-values dictionary
        # TODO: Is there a better way to do this so the length of the r?
        self.tree.seed_node._set_edge_length(0.0)
        _, _, p_val_dict = self._build_p_branch_length_values(
            self.tree.seed_node)

        # Initialize the matrix
        labels = self.pam.get_column_headers()
        mtx = Matrix(
            np.zeros((len(labels), len(p_val_dict)), dtype=float),
            headers={'0': labels, '1': list(p_val_dict.keys())})

        # We need a mapping of node path id to matrix column.  I don't think
        #     order matters
        node_col_idx = dict(
            zip(p_val_dict.keys(), range(len(p_val_dict.keys()))))

        for node_clade_id, p_val_node_val in p_val_dict.items():
            for tip_mtx_idx, tip_p_vals in p_val_node_val.items():
                mtx[int(tip_mtx_idx)][node_col_idx[node_clade_id]] = tip_p_vals

        return mtx



"""This module contains tools for transforming raw occurrence data."""
import csv
import sys

from lmpy.point import PointCsvReader


max_int = sys.maxsize
while True:
    # Decrease max_int value by factor of 10 if overflow
    try:
        csv.field_size_limit(max_int)
        break
    except OverflowError:  # pragma: no cover
        max_int = int(max_int / 10)


# .............................................................................
def get_chunk_key(value, group_position, group_size, filler_char='a'):
    """Get the chunk key from the value.

    Args:
        value (str): The value to pull the chunk key from.
        group_position (int): The position of the group chunk.
        group_size (int): The number of characters that comprises the group.
        filler_char (str): What to use for filler characters in case the group chunk is
            less than the desired number of characters.

    Returns:
        str: The group chunk key from the value.
    """
    value += 10 * filler_char
    return value[group_size * group_position:group_size * (group_position + 1)].lower()


# .............................................................................
def sort_points(readers, writer, wranglers=None):
    """Sort and clean occurrence points.

    Args:
        readers (:obj:`list` of :obj:`PointReader): A list of point readers.
        writer (PointWriter): A point writer to output points.
        wranglers (list of DataWranglers): An optional list of data wranglers
            to use to modify / filter points.
    """
    # Make sure readers is a list
    if isinstance(readers, PointCsvReader):
        readers = [readers]
    in_points = []
    # Sort points
    for reader in readers:
        for points in reader:
            in_points.extend(points)
    sorted_points = sorted(in_points)
    in_points = None
    # Filter if desired
    if wranglers:
        for flt in wranglers:
            if sorted_points:
                sorted_points = flt(sorted_points)
    # Write points
    writer.write_points(sorted_points)


# .............................................................................
def split_points(
    readers,
    writers,
    group_attribute,
    group_size,
    group_position,
    wranglers=None
):
    """Split points in the readers into smaller chunks for easier processing.

    Args:
        readers (:obj:`list` of :obj:`PointReader): A list of point readers.
        writers (dict): A dictionary of {combo: PointWriter} used for writing
            points.
        group_attribute (str): The Point attribute to be used to determine
            which group the Point belongs to.
        group_size (int): How many characters to use for determining group.
        group_position (int): Starting character index to use for determining
            group.
        wranglers (list of DataWranglers): An optional list of data wranglers
            to use to modify / filter points.

    Note:
        group_attribute should be used when instantiating the Point readers.
    """
    for reader in readers:
        for points in reader:
            # Assume all points returned go into the same group
            location_key = get_chunk_key(
                points[0].get_attribute(group_attribute), group_position,
                group_size)
            if wranglers:
                for flt in wranglers:
                    if points:
                        points = flt(points)
            writers[location_key].write_points(points)


# .............................................................................
def wrangle_points(readers, writer, wranglers=None):
    """Wrangle input points and write output points.

    Args:
        readers (:obj:`list` of :obj:`PointReader): A list of point readers.
        writer (PointWriter): A point writer to output points.
        wranglers (list of DataWranglers): An optional list of data wranglers
            to use to modify / filter points.
    """
    # Make sure readers is a list
    if isinstance(readers, PointCsvReader):
        readers = [readers]
    for reader in readers:
        for points in reader:
            if wranglers:
                for flt in wranglers:
                    if points:
                        points = flt(points)
            # Write wrangled points
            writer.write_points(points)


# .............................................................................
__all__ = ['get_chunk_key', 'sort_points', 'split_points', 'wrangle_points']



"""Module containing wrangler factory tools."""
from lmpy.data_wrangling.occurrence.filters import (
    get_attribute_filter, get_bounding_box_filter,
    get_decimal_precision_filter, get_disjoint_geometries_filter,
    get_intersect_geometries_filter, get_minimum_points_filter,
    get_spatial_index_filter, get_unique_localities_filter)
from lmpy.data_wrangling.occurrence.modifiers import (
    get_accepted_name_modifier, get_attribute_modifier,
    get_common_format_modifier, get_coordinate_converter_modifier)
from lmpy.spatial import SpatialIndex


# .............................................................................
class WRANGLER_TYPES:
    """Constants class for occurrence data wrangler types."""
    # Filters
    ATTRIBUTE_FILTER = 'attribute_filter'
    BBOX_FILTER = 'bbox_filter'
    DECIMAL_PRECISION_FILTER = 'decimal_precision_filter'
    DISJOINT_GEOMETRIES_FILTER = 'disjoint_geometries_filter'
    INTERSECT_GEOMETRIES_FILTER = 'intersect_geometries_filter'
    MINIMUM_POINTS_FILTER = 'minimum_points_filter'
    SPATIAL_INDEX_FILTER = 'spatial_index_filter'
    UNIQUE_LOCALITIES_FILTER = 'unique_localities_filter'
    # Modifiers
    ACCEPTED_NAME_MODIFIER = 'accepted_name_modifier'
    ATTRIBUTE_MODIFIER = 'attribute_modifier'
    ATTRIBUTE_MAP_MODIFIER = 'attribute_map_modifier'
    CORDINATE_CONVERTER_MODIFIER = 'coordinate_converter_modifier'


# .............................................................................
def wrangler_factory(wrangler_config):
    """Get an occurrence data wrangler from the wrangler configuration.

    Args:
        wrangler_config (dict): A dictionary of wrangler configuration parameters.

    Returns:
        Method: An occurrence data wrangler.
    """
    wrangler_type = wrangler_config['wrangler_type'].lower()
    occurrence_wrangler = None
    if wrangler_type == WRANGLER_TYPES.ATTRIBUTE_FILTER:
        att_name = wrangler_config['attribute_name']
        list_delimiter = wrangler_config['field_delimiter']
        bad_values = wrangler_config['condition']['all']['not_in']

        # Not in list
        def all_not_in(value):  # pragma: no cover
            field_values = value.split(list_delimiter)
            return all(
                [val.strip(']').strip('[').strip(
                    '"') not in bad_values for val in field_values])

        occurrence_wrangler = get_attribute_filter(att_name, all_not_in)
    elif wrangler_type == WRANGLER_TYPES.DECIMAL_PRECISION_FILTER:
        occurrence_wrangler = get_decimal_precision_filter(
            int(wrangler_config['decimal_precision']))
    elif wrangler_type == WRANGLER_TYPES.DISJOINT_GEOMETRIES_FILTER:
        # Get geometries (wkts)
        wkts = wrangler_config['geometry_wkts']
        occurrence_wrangler = get_disjoint_geometries_filter(wkts)
    elif wrangler_type == WRANGLER_TYPES.INTERSECT_GEOMETRIES_FILTER:
        wkts = wrangler_config['geometry_wkts']
        occurrence_wrangler = get_intersect_geometries_filter(wkts)
    elif wrangler_type == WRANGLER_TYPES.UNIQUE_LOCALITIES_FILTER:
        occurrence_wrangler = get_unique_localities_filter()
    elif wrangler_type == WRANGLER_TYPES.BBOX_FILTER:
        min_x = float(wrangler_config['min_x'])
        min_y = float(wrangler_config['min_y'])
        max_x = float(wrangler_config['max_x'])
        max_y = float(wrangler_config['max_y'])
        occurrence_wrangler = get_bounding_box_filter(
            min_x, min_y, max_x, max_y)
    elif wrangler_type == WRANGLER_TYPES.MINIMUM_POINTS_FILTER:
        occurrence_wrangler = get_minimum_points_filter(
            int(wrangler_config['minimum_points']))
    elif wrangler_type == WRANGLER_TYPES.SPATIAL_INDEX_FILTER:
        spatial_index = SpatialIndex(wrangler_config['index_file'])

        def check_hit_func(hit, check_vals):  # pragma: no cover
            for check_key, check_val in check_vals:
                if check_key in hit:
                    if hit[check_key] == check_val:
                        return True
            return False

        def get_valid_intersections_func(species_name):  # pragma: no cover
            ret_vals = []
            if species_name in wrangler_config['species']:
                for level, value in wrangler_config['species'][species_name]:
                    level_key = 'Level_{}_cod'.format(level)
                    if int(level) < 4:
                        level_key = level_key.upper()
                    ret_vals.append((level_key, value))
            return ret_vals

        occurrence_wrangler = get_spatial_index_filter(
            spatial_index, get_valid_intersections_func, check_hit_func)
    elif wrangler_type == WRANGLER_TYPES.ATTRIBUTE_MODIFIER:
        att_name = wrangler_config['attribute_name']
        map_dict = wrangler_config['map_values']

        def replace_func(in_value):  # pragma: no cover
            if in_value in map_dict:
                return map_dict[in_value]
            return None

        occurrence_wrangler = get_attribute_modifier(att_name, replace_func)
    elif wrangler_type == WRANGLER_TYPES.ATTRIBUTE_MAP_MODIFIER:
        mapping = wrangler_config['attribute_mapping']
        occurrence_wrangler = get_common_format_modifier(mapping)
    elif wrangler_type == WRANGLER_TYPES.ACCEPTED_NAME_MODIFIER:
        occurrence_wrangler = get_accepted_name_modifier(
            wrangler_config['filename'])
    elif wrangler_type == WRANGLER_TYPES.CORDINATE_CONVERTER_MODIFIER:
        in_epsg = wrangler_config['in_epsg']
        out_epsg = wrangler_config['out_epsg']
        occurrence_wrangler = get_coordinate_converter_modifier(
            in_epsg, out_epsg)
    return occurrence_wrangler



"""Module containing occurrence data wranglers for modifying point data."""
from copy import deepcopy

# import osgeo
from osgeo import osr

from lmpy.point import Point
from lmpy.data_wrangling.occurrence.common import get_occurrence_map


# .............................................................................
def get_accepted_name_modifier(accepted_taxa_filename):
    """Get a data wrangler for populating accepted taxon name for points.

    Args:
        accepted_taxa_filename (str): A file containing accepted taxon names.

    Returns:
        Method: A function for filling in accepted taxon name for points.
    """
    accepted_map = {}
    with open(accepted_taxa_filename, 'r') as in_file:
        _ = next(in_file)
        for line in in_file:
            parts = line.split(',')
            accepted_map[parts[0].strip()] = parts[1].strip()

    # .......................
    def accepted_taxon_wrangler(points):
        return_points = []
        if isinstance(points, Point):
            points = [points]
        for point in points:
            check_sp = point.species_name
            # Check if we don't have an entry for this name
            if check_sp not in accepted_map.keys():
                # We don't have it, it doesn't exist
                new_name = ''
                accepted_map[check_sp] = new_name
            accepted_taxon_name = accepted_map[check_sp]
            if accepted_taxon_name is None or len(accepted_taxon_name) < 2:
                # print('No accepted name for {}'.format(check_sp))
                pass
            else:
                point.species_name = accepted_taxon_name
                return_points.append(point)
        return return_points

    return accepted_taxon_wrangler


# .............................................................................
def get_attribute_modifier(att_name, replace_func):
    """Get a function that replaces an Point attribute using a replace func.

    Args:
        att_name (str): The point attribute to modify.
        replace_func (method): A method that returns a replacement value.

    Returns:
        Method: A function that takes a list of points and returns points with the
            coordinates converted.
    """
    # .......................
    def modify_func(point):
        old_val = point.get_attribute(att_name)
        new_pt = deepcopy(point)
        new_pt.set_attribute(att_name, replace_func(old_val))
        return new_pt

    return get_occurrence_map(modify_func)


# .............................................................................
def get_common_format_modifier(attribute_map):
    """Get a function to convert points to a common format (common attributes).

    Args:
        attribute_map (dict): A map of current attribute name to new attribute
            name.

    Returns:
        Method: A function that takes a list of points and returns points with the
            coordinates converted.
    """
    # .......................
    def modify_func(point):
        new_pt = Point(
            point.species_name, point.x, point.y,
            attributes={
                new_name: point.get_attribute(old_name)
                for old_name, new_name in attribute_map.items()})
        return new_pt

    return get_occurrence_map(modify_func)


# .............................................................................
def get_coordinate_converter_modifier(in_epsg, out_epsg):
    """Get a function to convert points from one coordinate system to another.

    Args:
        in_epsg (int): The EPSG code of the incoming points.
        out_epsg (int): The target EPSG code for output points.

    Returns:
        Method: A function that takes a list of points and returns points with the
            coordinates converted.
    """
    source = osr.SpatialReference()
    source.ImportFromEPSG(in_epsg)

    # TODO: Change this in the future when we require GDAL >= 3
    # if int(osgeo.__version__[0]) >= 3:  # pragma: no cover
    #    # GDAL 3 changes axis order: https://github.com/OSGeo/gdal/issues/1546
    #    source.SetAxisMappingStrategy(osr.OAMS_TRADITIONAL_GIS_ORDER)

    target = osr.SpatialReference()
    target.ImportFromEPSG(out_epsg)

    transform = osr.CoordinateTransformation(source, target)

    # .......................
    def converter_func(point):
        new_x, new_y, _ = transform.TransformPoint(point.x, point.y)
        pt = deepcopy(point)
        pt.x = new_x
        pt.y = new_y
        return pt

    return get_occurrence_map(converter_func)



"""Module containing common functions for occurrence data wranglers."""
from lmpy import Point


# .............................................................................
def make_list(points):
    """Make sure that 'points' is a list.

    Args:
        points (list or Point): A point or list of points.

    Returns:
        list of Point: A list of Point objects.
    """
    if isinstance(points, Point):
        return [points]
    return points


# .............................................................................
def get_occurrence_filter(pass_condition_func):
    """Function to get a filter from a pass condition function.

    Args:
        pass_condition_func (Method): A function for determining if a Point should be
            filtered.

    Returns:
        Method: A function to filter a list of points.
    """
    # .......................
    def occurrence_filter(points):
        return list(filter(pass_condition_func, points))
    return occurrence_filter


# .............................................................................
def get_occurrence_map(map_func):
    """Function to get a mapper to apply a function to all points.

    Args:
        map_func (Method): A function to apply to all points.

    Returns:
        Method: A function to map the provied function to all points.
    """
    # .......................
    def occurrence_mapper(points):
        return list(map(map_func, make_list(points)))

    return occurrence_mapper
